/* 
   Filename: qzcomplex.cc

   Octave function to compute a complex generalized Schur (QZ) decomposition.
   This function is simply a wrapper function for the Fortran LAPACK function zgges.f
   
   Inputs: complex matricies A and B
   
   Outputs: complex matricies AA,BB,Q,Z such that:
   Q'*Q = I, Z'*Z = I,
   A = Q*AA*Z'  (or equivalently Q'*A*Z = AA)
   B = Q*BB*Z'  (or equivalently Q'*B*Z = BB)
*/

#include <octave/oct.h>
#include "f77-fcn.h"

typedef struct{double real, imag;} doublecomplex;

extern "C"
{
  int F77_FUNC(zgges, ZGGES) (char* JOBCSL, char* JOBVSR, char* SORT, int *DELCTG, int *N,
			      doublecomplex *A, int* LDA, 
			      doublecomplex *B, int* LDB, int* SDIM, 
			      doublecomplex *ALPHA, doublecomplex *BETA,
			      doublecomplex *VSL, int *LDVSL, 
			      doublecomplex *VSR, int *LDVSR, 
			      doublecomplex *WORK, int *LWORK, double *RWORK, 
			      int *BWORK, int *INFO);
}

doublecomplex* createmat(int rows, int cols);
void deletemat(doublecomplex *A, int rows);
doublecomplex* assign_dc_mat(ComplexMatrix A);
ComplexMatrix assign_mat_dc(doublecomplex *A, int rs, int cs);

//int main()
DEFUN_DLD(qzcomplex, args, , "[AA,BB,Q,Z] = qzcomplex(A,B)\nOctave wrapper for the Fortran LAPACK function zgges.f\nComputes complex QZ decomposition where\nQ'*A*Z = AA, Q'*B*Z = BB, Q'*Q = I, Z'*Z = I")
{
  octave_value_list returnval;
 
  ComplexMatrix Ao = args(0).complex_matrix_value();
  ComplexMatrix Bo = args(1).complex_matrix_value();
  int lda = Ao.rows();
  int n = Ao.cols();
  int ldb = Bo.rows();

  doublecomplex *A = assign_dc_mat(Ao); 
  doublecomplex *B = assign_dc_mat(Bo); 
    
  // outputs from zgges.f
  int ldvsl = n; // dimension of Q
  int ldvsr = n; // dimestion of Z
  doublecomplex *alpha = new doublecomplex[n];
  doublecomplex *beta = new doublecomplex[n];
  doublecomplex *Q = createmat(ldvsl,n);
  doublecomplex *Z = createmat(ldvsr,n);
  int sdim; // output from zgges.f, meaningless since sort = 'N'.
  int info; 

  char jobvsl = 'V';
  char jobvsr = 'V';
  char sort = 'N';
  int delctg = 0; // Not referenced since sort = 'N'.

  int lwork = 33*n;
  doublecomplex *work = new doublecomplex[lwork];
  double* rwork = new double[8*n];

  int bwork = 0; // Not referenced since sort = 'N'.
  
  F77_XFCN(zgges, ZGGES, (&jobvsl, &jobvsr, &sort, &delctg, &n, A, &lda, B, &ldb, &sdim, alpha, beta, Q, &ldvsl, Z, &ldvsr, work, &lwork, rwork, &bwork, &info) );

 
  ComplexMatrix AAo = assign_mat_dc(A,n,n);
  ComplexMatrix BBo = assign_mat_dc(B,n,n);
  ComplexMatrix Qo = assign_mat_dc(Q,n,n);
  ComplexMatrix Zo = assign_mat_dc(Z,n,n);

  // Clean up
  deletemat(A,lda);
  deletemat(B,ldb);
  delete [] alpha;
  delete [] beta;
  deletemat(Q,ldvsl);
  deletemat(Z,ldvsr);
  delete [] work;
  delete [] rwork;

  // Return octave values.
  returnval(0) = AAo;
  returnval(1) = BBo;
  returnval(2) = Qo;
  returnval(3) = Zo;

  return returnval;
}

doublecomplex* createmat(int rows, int cols)
{
  int size = rows*cols;
  doublecomplex* A = new doublecomplex[size];
  return A;
}

void deletemat(doublecomplex *A, int rows)
{
  delete [] A;
}

doublecomplex* assign_dc_mat(ComplexMatrix A)
{
  int rows = A.rows();
  int cols = A.cols();

  doublecomplex *B = createmat(rows,cols);
  int index = 0;
  for(int c=0;c<cols;c++) {
    for(int r=0;r<rows;r++) {
      B[index].real = real(A(r,c));
      B[index].imag = imag(A(r,c));
      index++;
    }
  }

  return B;
}

ComplexMatrix assign_mat_dc(doublecomplex *A, int rows, int cols)
{
  ComplexMatrix B = ComplexMatrix(rows,cols);
  
  int index=0;
  for(int c=0; c<cols; c++) {
    for(int r=0; r<rows; r++) {
      B(r,c) = Complex(A[index].real, A[index].imag);
      index++;
    }
  }

  return B;
}
